import unittest

import claripy


class TestFallbackAbstraction(unittest.TestCase):
    def test_fallback_abstraction(self):
        bz = claripy.backends.z3

        a = claripy.BVV(5, 32)
        b = claripy.BVS("x", 32, explicit_name=True)
        c = a + b
        d = 5 + b
        e = b + 5
        f = b + b
        g = a + a

        assert not a.symbolic
        assert b.symbolic
        assert c.symbolic
        assert d.symbolic
        assert e.symbolic
        assert f.symbolic

        assert type(claripy.backends.concrete.convert(a)) is claripy.bv.BVV
        assert type(claripy.backends.concrete.convert(g)) is claripy.bv.BVV

        self.assertRaises(claripy.errors.BackendError, claripy.backends.concrete.convert, b)
        self.assertRaises(claripy.errors.BackendError, claripy.backends.concrete.convert, c)
        self.assertRaises(claripy.errors.BackendError, claripy.backends.concrete.convert, d)
        self.assertRaises(claripy.errors.BackendError, claripy.backends.concrete.convert, e)
        self.assertRaises(claripy.errors.BackendError, claripy.backends.concrete.convert, f)

        assert str(bz.convert(b)) == "x"
        assert bz.convert(b).__module__ == "z3.z3"

        assert str(bz.convert(c)) == "5 + x"
        assert str(bz.convert(d)) == "5 + x"
        assert str(bz.convert(e)) == "x + 5"
        assert str(bz.convert(f)) == "x + x"


if __name__ == "__main__":
    unittest.main()



# pylint:disable=missing-class-docstring,multiple-statements
import claripy


class AnnotationA(claripy.Annotation):
    def __init__(self, letter, number):
        self.letter = letter
        self.number = number
        claripy.Annotation.__init__(self)


class AnnotationB(AnnotationA):
    @property
    def eliminatable(self):
        return False

    @property
    def relocatable(self):
        return False


class AnnotationC(AnnotationA):
    def __init__(self, letter, number):
        super().__init__(letter, number)
        self._relocatable = True

    @property
    def eliminatable(self):
        return False

    @property
    def relocatable(self):
        return self._relocatable

    def relocate(self, src, dst):
        return AnnotationC(self.letter, self.number + 1)


class AnnotationD(claripy.Annotation):
    @property
    def eliminatable(self):
        return False

    @property
    def relocatable(self):
        return True


class BackendA(claripy.Backend):
    def __init__(self):
        claripy.Backend.__init__(self)
        self._op_expr["BVV"] = lambda e, result=None: e.args[0]

    def apply_annotation(self, o, a):
        return o + a.number


def test_backend():
    x = claripy.BVV(10, 32).annotate(AnnotationA("a", 1))
    assert BackendA().convert(x) == 11


def test_simplification():
    x = claripy.BVS("x", 32).annotate(AnnotationA("a", 1))
    y = x ^ x
    assert y.depth == 1
    assert len(y.annotations) == 0

    x = claripy.BVS("x", 32).annotate(AnnotationB("a", 1))
    y = x ^ x
    assert y.depth == 2

    x = claripy.BVS("x", 32).annotate(AnnotationC("a", 1))
    y = x ^ x
    assert y.depth == 1
    assert len(y.annotations) == 1
    assert y.annotations[0].number == 2


def test_missing_annotations_from_simplification():
    relocatable_anno = AnnotationC("a", 2)

    x0 = claripy.BVS("x", 32)
    x1 = claripy.BVV(24, 32)
    k = (x1 + x0).annotate(relocatable_anno)

    x3 = claripy.simplify(k)

    assert len(x3.annotations) == 1


def test_annotations():
    x = claripy.BVS("x", 32) + 1
    xx = x._apply_to_annotations(lambda a: a)
    assert x is xx

    a1 = AnnotationA("a", 1)
    a2 = AnnotationA("a", 1)

    x1 = x.annotate(a1)
    x2 = x1.annotate(a2)
    x2a = x.annotate(a1, a2)
    x3 = x2.remove_annotation(a1)
    x4 = x3.remove_annotation(a2)
    x5 = x2.remove_annotations({a1, a2})

    assert x.variables == x1.variables
    assert x.variables == x2.variables
    assert x.variables == x2a.variables
    assert x.variables == x3.variables
    assert x.variables == x4.variables
    assert x.variables == x5.variables

    assert x is not x1
    assert x is not x2
    assert x is not x3
    assert x1 is not x2
    assert x1 is not x3
    assert x2 is not x3
    assert x2 is x2a
    assert x is x4
    assert x is x5

    assert x.op == x1.op
    assert x.annotations == ()
    assert x1.annotations == (a1,)
    assert x2.annotations == (a1, a2)
    assert x3.annotations == (a2,)

    assert claripy.backends.z3.convert(x).eq(claripy.backends.z3.convert(x3))

    const = claripy.BVV(1, 32)
    consta = const.annotate(AnnotationB("a", 0))
    const1 = consta + 1
    const1a = const1.annotate(AnnotationB("b", 1))
    const2 = const1a + 1
    # const2 should be (const1a + 1), instead of (1 + 1 + 1)
    # the flatten simplifier for __add__ should not be applied as AnnotationB is not relocatable (and not eliminatable)
    assert const2.depth == 3


def test_eagerness():
    x = claripy.BVV(10, 32).annotate(AnnotationD())
    y = x + 1
    assert y.annotations == x.annotations


def test_ast_hash_should_consider_relocatable_annotations():
    relocatable_anno = AnnotationC("a", 2)
    const = claripy.BVV(1337, 32)
    x0 = claripy.BVS("x", 32).annotate(relocatable_anno)
    y0 = claripy.Concat(x0, const)

    # make the annotation not relocatable
    # this is of course a hack, but it can demonstrate the problem
    relocatable_anno._relocatable = False
    x0._relocatable_annotations = frozenset()

    y1 = claripy.Concat(x0, const)

    assert len(y0.annotations) == 1
    assert len(y1.annotations) == 0
    assert y0._hash != y1._hash


def test_remove_relocatable_annotations():
    relocatable_anno = AnnotationC("a", 2)
    const = claripy.BVV(1337, 32)

    x0 = claripy.BVS("x", 32).annotate(relocatable_anno)
    y0 = claripy.Concat(x0, const)
    assert len(y0.annotations) == 1
    assert y0.annotations == (relocatable_anno,)

    y1 = y0.remove_annotation(relocatable_anno)

    assert len(y1.annotations) == 0


def test_duplicated_annotations_from_makelike():
    relocatable_anno = AnnotationC("a", 2)

    x0 = claripy.BVS("x", 32).annotate(relocatable_anno)
    x1 = claripy.BVV(24, 32)

    # make_like() should not re-apply child annotations if the child is the make_like target
    x2 = x0 + x1
    assert len(x2.annotations) == 1

    # simplify() should not re-apply annotations since annotations are kept during the simplification process by
    # make_like().
    x3 = claripy.simplify(x0 + x1)
    assert len(x3.annotations) == 1


if __name__ == "__main__":
    test_annotations()
    test_backend()
    test_eagerness()
    test_ast_hash_should_consider_relocatable_annotations()
    test_remove_relocatable_annotations()
    test_duplicated_annotations_from_makelike()
    test_simplification()



import unittest
import claripy

from claripy import frontend_mixins, backend_manager
from claripy.backends.backend_smtlib import BackendSMTLibBase
from claripy.frontends.constrained_frontend import ConstrainedFrontend
from claripy.ast.strings import String

KEEP_TEST_PERFORMANT = True


class TestSMTLibBackend(unittest.TestCase):
    def get_solver(self):
        backend_manager.backends._register_backend(BackendSMTLibBase(), "smt", False, False)

        class SolverSMT(
            frontend_mixins.ConstraintFixerMixin,
            frontend_mixins.ConcreteHandlerMixin,
            frontend_mixins.ConstraintFilterMixin,
            frontend_mixins.ConstraintDeduplicatorMixin,
            frontend_mixins.EagerResolutionMixin,
            frontend_mixins.SMTLibScriptDumperMixin,
            ConstrainedFrontend,
        ):  # pylint:disable=abstract-method
            def __init__(self, *args, **kwargs):
                self._solver_backend = backend_manager.backends.smt
                super().__init__(*args, **kwargs)

        return SolverSMT()

    def test_concat(self):
        correct_script = """(set-logic ALL)
(declare-fun {0}symb_concat () String)
(assert (let ((.def_0 (str.++  "conc" {0}symb_concat))) (let ((.def_1 (= .def_0 "concrete"))) .def_1)))
(check-sat)
""".format(
            String.STRING_TYPE_IDENTIFIER
        )
        str_concrete = claripy.StringV("conc")
        str_symbol = claripy.StringS("symb_concat", 4, explicit_name=True)
        solver = self.get_solver()
        res = str_concrete + str_symbol
        solver.add(res == claripy.StringV("concrete"))
        script = solver.get_smtlib_script_satisfiability()
        # with open("dump_concat.smt2", "w") as dump_f:
        # dump_f.write(script)
        self.assertEqual(correct_script, script)

    def test_concat_simplification(self):
        correct_script = """(set-logic ALL)


(check-sat)
"""
        solver = self.get_solver()
        str_concrete = claripy.StringV("conc")
        res = str_concrete + str_concrete + str_concrete
        res2 = claripy.StrConcat(str_concrete, str_concrete) + str_concrete
        solver.add(res == res2)
        script = solver.get_smtlib_script_satisfiability()
        self.assertEqual(correct_script, script)

    def test_substr(self):
        correct_script = """(set-logic ALL)
(declare-fun {0}symb_subst () String)
(assert (let ((.def_0 (= ( str.substr {0}symb_subst 1 2) "on"))) .def_0))
(check-sat)
""".format(
            String.STRING_TYPE_IDENTIFIER
        )
        str_symbol = claripy.StringS("symb_subst", 4, explicit_name=True)
        solver = self.get_solver()
        res = claripy.StrSubstr(1, 2, str_symbol) == claripy.StringV("on")
        solver.add(res)
        script = solver.get_smtlib_script_satisfiability()
        # with open("dump_substr.smt2", "w") as dump_f:
        # dump_f.write(script)
        self.assertEqual(correct_script, script)

    def test_substr_BV_concrete_index(self):
        correct_script = """(set-logic ALL)
(declare-fun {0}symb_subst () String)
(assert (let ((.def_0 (= ( str.substr {0}symb_subst 1 2) "on"))) .def_0))
(check-sat)
""".format(
            String.STRING_TYPE_IDENTIFIER
        )
        str_symbol = claripy.StringS("symb_subst", 4, explicit_name=True)
        solver = self.get_solver()
        bv1 = claripy.BVV(1, 32)
        bv2 = claripy.BVV(2, 32)
        res = claripy.StrSubstr(bv1, bv2, str_symbol) == claripy.StringV("on")
        solver.add(res)
        script = solver.get_smtlib_script_satisfiability()
        # with open("dump_substr_bv_concrete.smt2", "w") as dump_f:
        # dump_f.write(script)
        self.assertEqual(correct_script, script)

    def test_substr_BV_symbolic_index(self):
        correct_script = """(set-logic ALL)
(declare-fun {0}symb_subst () String)
(declare-fun symb_subst_count () Int)
(declare-fun symb_subst_start_idx () Int)
(assert (let ((.def_0 (= ( str.substr {0}symb_subst symb_subst_start_idx symb_subst_count) "on"))) .def_0))
(check-sat)
""".format(
            String.STRING_TYPE_IDENTIFIER
        )
        str_symbol = claripy.StringS("symb_subst", 4, explicit_name=True)
        solver = self.get_solver()
        bv1 = claripy.BVS("symb_subst_start_idx", 32, explicit_name=True)
        bv2 = claripy.BVS("symb_subst_count", 32, explicit_name=True)
        res = claripy.StrSubstr(bv1, bv2, str_symbol) == claripy.StringV("on")
        solver.add(res)
        script = solver.get_smtlib_script_satisfiability()
        # with open("dump_substr_bv_symbolic.smt2", "w") as dump_f:
        #     dump_f.write(script)
        self.assertEqual(correct_script, script)

    def test_substr_BV_mixed_index(self):
        correct_script = """(set-logic ALL)
(declare-fun {0}symb_subst () String)
(declare-fun symb_subst_start_idx () Int)
(assert (let ((.def_0 (= ( str.substr {0}symb_subst symb_subst_start_idx 2) "on"))) .def_0))
(check-sat)
""".format(
            String.STRING_TYPE_IDENTIFIER
        )
        str_symbol = claripy.StringS("symb_subst", 4, explicit_name=True)
        solver = self.get_solver()
        bv1 = claripy.BVS("symb_subst_start_idx", 32, explicit_name=True)
        bv2 = claripy.BVV(2, 32)
        res = claripy.StrSubstr(bv1, bv2, str_symbol) == claripy.StringV("on")
        solver.add(res)
        script = solver.get_smtlib_script_satisfiability()
        # with open("dump_substr_bv_symbolic.smt2", "w") as dump_f:
        #     dump_f.write(script)
        self.assertEqual(correct_script, script)

    def test_substr_simplification(self):
        correct_script = """(set-logic ALL)


(check-sat)
"""
        str_concrete = claripy.StringV("concrete")
        solver = self.get_solver()
        solver.add(claripy.StrSubstr(1, 2, str_concrete) == claripy.StringV("on"))
        script = solver.get_smtlib_script_satisfiability()
        self.assertEqual(correct_script, script)

    def test_replace(self):
        correct_script = """(set-logic ALL)
(declare-fun {0}symb_repl () String)
(assert (let ((.def_0 (= ( str.replace {0}symb_repl "a" "b" ) "cbne"))) .def_0))
(check-sat)
""".format(
            String.STRING_TYPE_IDENTIFIER
        )
        str_to_replace_symb = claripy.StringS("symb_repl", 4, explicit_name=True)
        sub_str_to_repl = claripy.StringV("a")
        replacement = claripy.StringV("b")
        solver = self.get_solver()
        repl_stringa = claripy.StrReplace(str_to_replace_symb, sub_str_to_repl, replacement)
        solver.add(repl_stringa == claripy.StringV("cbne"))
        script = solver.get_smtlib_script_satisfiability()
        # with open("dump_replace.smt2", "w") as dump_f:
        # dump_f.write(script)
        self.assertEqual(correct_script, script)

    def test_replace_simplification(self):
        correct_script = """(set-logic ALL)


(check-sat)
"""
        str_to_replace = claripy.StringV("cane")
        sub_str_to_repl = claripy.StringV("a")
        replacement = claripy.StringV("b")
        repl_stringa = claripy.StrReplace(str_to_replace, sub_str_to_repl, replacement)
        solver = self.get_solver()
        solver.add(repl_stringa == claripy.StringV("cbne"))
        solver = self.get_solver()
        script = solver.get_smtlib_script_satisfiability()
        self.assertEqual(correct_script, script)

    def test_ne(self):
        correct_script = """(set-logic ALL)
(declare-fun {0}symb_ne () String)
(assert (let ((.def_0 (= {0}symb_ne "concrete"))) (let ((.def_1 (not .def_0))) .def_1)))
(check-sat)
""".format(
            String.STRING_TYPE_IDENTIFIER
        )
        str_symb = claripy.StringS("symb_ne", 12, explicit_name=True)
        solver = self.get_solver()
        solver.add(str_symb != claripy.StringV("concrete"))
        script = solver.get_smtlib_script_satisfiability()
        # with open("dump_ne.smt2", "w") as dump_f:
        #     dump_f.write(script)
        self.assertEqual(correct_script, script)

    def test_length(self):
        correct_script = """(set-logic ALL)
(declare-fun {0}symb_length () String)
(assert (let ((.def_0 (= (str.len {0}symb_length) 14))) .def_0))
(check-sat)
""".format(
            String.STRING_TYPE_IDENTIFIER
        )
        str_symb = claripy.StringS("symb_length", 12, explicit_name=True)
        solver = self.get_solver()
        # TODO: How do we want to dela with the size of a symbolic string?
        solver.add(claripy.StrLen(str_symb, 32) == 14)
        script = solver.get_smtlib_script_satisfiability()
        # with open("dump_length.smt2", "w") as dump_f:
        #     dump_f.write(script)
        self.assertEqual(correct_script, script)

    def test_length_simplification(self):
        correct_script = """(set-logic ALL)


(check-sat)
"""
        str_concrete = claripy.StringV("concrete")
        solver = self.get_solver()
        solver.add(claripy.StrLen(str_concrete, 32) == 8)
        script = solver.get_smtlib_script_satisfiability()
        self.assertEqual(correct_script, script)

    def test_or(self):
        correct_script = """(set-logic ALL)
(declare-fun {0}Symb_or () String)
(assert (let ((.def_0 (= {0}Symb_or "ciao"))) (let ((.def_1 (= {0}Symb_or "abc"))) (let ((.def_2 (or .def_1 .def_0))) .def_2))))
(check-sat)
""".format(
            String.STRING_TYPE_IDENTIFIER
        )
        str_symb = claripy.StringS("Symb_or", 4, explicit_name=True)
        solver = self.get_solver()
        res = claripy.Or((str_symb == claripy.StringV("abc")), (str_symb == claripy.StringV("ciao")))
        solver.add(res)
        script = solver.get_smtlib_script_satisfiability()
        # with open("dump_or.smt2", "w") as dump_f:
        #     dump_f.write(script)
        self.assertEqual(correct_script, script)

    def test_lt_etc(self):
        correct_script = """(set-logic ALL)
(declare-fun {0}Symb_lt_test () String)
(assert (let ((.def_0 (<= (str.len {0}Symb_lt_test) 4))) .def_0))
(assert (let ((.def_0 (< (str.len {0}Symb_lt_test) 4))) .def_0))
(assert (let ((.def_0 (<= 4 (str.len {0}Symb_lt_test)))) .def_0))
(assert (let ((.def_0 (< 4 (str.len {0}Symb_lt_test)))) .def_0))
(check-sat)
""".format(
            String.STRING_TYPE_IDENTIFIER
        )
        str_symb = claripy.StringS("Symb_lt_test", 4, explicit_name=True)
        solver = self.get_solver()
        c1 = claripy.StrLen(str_symb, 32) <= 4
        c2 = claripy.StrLen(str_symb, 32) < 4
        c3 = claripy.StrLen(str_symb, 32) >= 4
        c4 = claripy.StrLen(str_symb, 32) > 4
        solver.add(c1)
        solver.add(c2)
        solver.add(c3)
        solver.add(c4)
        script = solver.get_smtlib_script_satisfiability()
        # with open("dump_lt_etc.smt2", "w") as dump_f:
        #     dump_f.write(script)
        self.assertEqual(correct_script, script)

    def test_contains(self):
        correct_script = """(set-logic ALL)
(declare-fun {0}symb_contains () String)
(assert ( str.contains {0}symb_contains "an"))
(check-sat)
""".format(
            String.STRING_TYPE_IDENTIFIER
        )
        str_symb = claripy.StringS("symb_contains", 4, explicit_name=True)
        res = claripy.StrContains(str_symb, claripy.StringV("an"))
        solver = self.get_solver()
        solver.add(res)
        script = solver.get_smtlib_script_satisfiability()
        # with open("dump_contains.smt2", "w") as dump_f:
        #     dump_f.write(script)
        self.assertEqual(correct_script, script)

    def test_contains_simplification(self):
        correct_script = """(set-logic ALL)


(check-sat)
"""
        str_concrete = claripy.StringV("concrete")
        solver = self.get_solver()
        res = claripy.StrContains(str_concrete, claripy.StringV("nc"))
        solver.add(res)
        script = solver.get_smtlib_script_satisfiability()
        self.assertEqual(correct_script, script)

    def test_prefix(self):
        correct_script = """(set-logic ALL)
(declare-fun {0}symb_prefix () String)
(assert ( str.prefixof "an" {0}symb_prefix ))
(check-sat)
""".format(
            String.STRING_TYPE_IDENTIFIER
        )
        str_symb = claripy.StringS("symb_prefix", 4, explicit_name=True)
        res = claripy.StrPrefixOf(claripy.StringV("an"), str_symb)
        solver = self.get_solver()
        solver.add(res)
        script = solver.get_smtlib_script_satisfiability()
        # with open("dump_prefix.smt2", "w") as dump_f:
        #     dump_f.write(script)
        self.assertEqual(correct_script, script)

    def test_suffix(self):
        correct_script = """(set-logic ALL)
(declare-fun {0}symb_suffix () String)
(assert ( str.suffixof "an" {0}symb_suffix ))
(check-sat)
""".format(
            String.STRING_TYPE_IDENTIFIER
        )
        str_symb = claripy.StringS("symb_suffix", 4, explicit_name=True)
        res = claripy.StrSuffixOf(claripy.StringV("an"), str_symb)
        solver = self.get_solver()
        solver.add(res)
        script = solver.get_smtlib_script_satisfiability()
        # with open("dump_suffix.smt2", "w") as dump_f:
        #     dump_f.write(script)
        self.assertEqual(correct_script, script)

    def test_prefix_simplification(self):
        correct_script = """(set-logic ALL)


(check-sat)
"""
        str_concrete = claripy.StringV("concrete")
        solver = self.get_solver()
        res = claripy.StrPrefixOf(claripy.StringV("conc"), str_concrete)
        solver.add(res)
        script = solver.get_smtlib_script_satisfiability()
        self.assertEqual(correct_script, script)

    def test_suffix_simplification(self):
        correct_script = """(set-logic ALL)


(check-sat)
"""
        str_concrete = claripy.StringV("concrete")
        solver = self.get_solver()
        res = claripy.StrSuffixOf(claripy.StringV("rete"), str_concrete)
        solver.add(res)
        script = solver.get_smtlib_script_satisfiability()
        self.assertEqual(correct_script, script)

    def test_index_of(self):
        correct_script = """(set-logic ALL)
(declare-fun {0}symb_suffix () String)
(assert ( str.indexof {0}symb_suffix "an" 0 ))
(check-sat)
""".format(
            String.STRING_TYPE_IDENTIFIER
        )
        str_symb = claripy.StringS("symb_suffix", 4, explicit_name=True)
        res = claripy.StrIndexOf(str_symb, claripy.StringV("an"), 0, 32)
        solver = self.get_solver()
        solver.add(res)
        script = solver.get_smtlib_script_satisfiability()
        # with open("dump_suffix.smt2", "w") as dump_f:
        #     dump_f.write(script)
        self.assertEqual(correct_script, script)

    def test_index_of_simplification(self):
        correct_script = """(set-logic ALL)


(check-sat)
"""
        str_concrete = claripy.StringV("concrete")
        solver = self.get_solver()
        res = claripy.StrIndexOf(str_concrete, claripy.StringV("rete"), 0, 32)
        solver.add(res == 4)
        script = solver.get_smtlib_script_satisfiability()
        self.assertEqual(correct_script, script)

    def test_index_of_symbolic_start_idx(self):
        correct_script = """(set-logic ALL)
(declare-fun {0}symb_index_of () String)
(declare-fun symb_start_idx () Int)
(assert (let ((.def_0 (< 32 symb_start_idx))) .def_0))
(assert (let ((.def_0 (< symb_start_idx 35))) .def_0))
(assert (let ((.def_0 (= ( str.indexof {0}symb_index_of "an" symb_start_idx ) 33))) .def_0))
(check-sat)
""".format(
            String.STRING_TYPE_IDENTIFIER
        )
        str_symb = claripy.StringS("symb_index_of", 4, explicit_name=True)
        start_idx = claripy.BVS("symb_start_idx", 32, explicit_name=True)

        solver = self.get_solver()

        solver.add(start_idx > 32)
        solver.add(start_idx < 35)
        res = claripy.StrIndexOf(str_symb, claripy.StringV("an"), start_idx, 32)

        solver.add(res == 33)
        script = solver.get_smtlib_script_satisfiability()
        # with open("dump_suffix.smt2", "w") as dump_f:
        #     dump_f.write(script)
        self.assertEqual(correct_script, script)

    def test_str_to_int(self):
        correct_script = """(set-logic ALL)
(declare-fun {0}symb_strtoint () String)
(assert (let ((.def_0 (= ( str.to.int {0}symb_strtoint ) 12))) .def_0))
(check-sat)
""".format(
            String.STRING_TYPE_IDENTIFIER
        )
        str_symb = claripy.StringS("symb_strtoint", 4, explicit_name=True)
        res = claripy.StrToInt(str_symb, 32)
        solver = self.get_solver()
        solver.add(res == 12)
        script = solver.get_smtlib_script_satisfiability()
        # with open("dump_strtoint.smt2", "w") as dump_f:
        #     dump_f.write(script)
        self.assertEqual(correct_script, script)

    def test_str_to_int_simplification(self):
        correct_script = """(set-logic ALL)


(check-sat)
"""
        str_concrete = claripy.StringV("12")
        solver = self.get_solver()
        res = claripy.StrToInt(str_concrete, 32)
        solver.add(res == 12)
        script = solver.get_smtlib_script_satisfiability()
        self.assertEqual(correct_script, script)

    def test_is_digit(self):
        correct_script = """(set-logic ALL)
(declare-fun STRING_symb_str_is_digit () String)
(assert (let ((.def_0 (= ( str.to.int STRING_symb_str_is_digit ) (- 1)))) (let ((.def_1 (not .def_0))) .def_1)))
(check-sat)
"""
        str_symb = claripy.StringS("symb_str_is_digit", 12, explicit_name=True)
        res = claripy.StrIsDigit(str_symb)
        solver = self.get_solver()
        solver.add(res)
        script = solver.get_smtlib_script_satisfiability()
        self.assertEqual(correct_script, script)

    def test_is_digit_simplification(self):
        correct_script = """(set-logic ALL)


(check-sat)
"""
        str_concrete = claripy.StringV("1")
        res = claripy.StrIsDigit(str_concrete)
        solver = self.get_solver()
        solver.add(res)
        script = solver.get_smtlib_script_satisfiability()
        self.assertEqual(correct_script, script)

    def test_int_to_str(self):
        correct_script = """(set-logic ALL)
(declare-fun symb_int () Int)
(assert (let ((.def_0 (= ( int.to.str symb_int ) "12"))) .def_0))
(check-sat)
"""
        int_symb = claripy.BVS("symb_int", 32, explicit_name=True)
        res = claripy.IntToStr(int_symb)
        solver = self.get_solver()
        solver.add(res == claripy.StringV("12"))
        script = solver.get_smtlib_script_satisfiability()
        self.assertEqual(correct_script, script)

    def test_int_to_str_simplification(self):
        correct_script = """(set-logic ALL)


(check-sat)
"""
        int_concrete = claripy.BVV(0xC, 32)
        res = claripy.IntToStr(int_concrete)
        solver = self.get_solver()
        solver.add(res == claripy.StringV("12"))
        script = solver.get_smtlib_script_satisfiability()
        self.assertEqual(correct_script, script)


if __name__ == "__main__":
    suite = unittest.TestLoader().loadTestsFromTestCase(TestSMTLibBackend)
    unittest.TextTestRunner(verbosity=2).run(suite)



import claripy


def test_bool_simplification():
    def assert_correct(a, b):
        assert claripy.backends.z3.identical(claripy.simplify(a), b)

    a, b, c = (claripy.BoolS(name) for name in ("a", "b", "c"))

    assert_correct(claripy.And(a, claripy.Not(a)), claripy.false)
    assert_correct(claripy.Or(a, claripy.Not(a)), claripy.true)

    complex_true_expression = claripy.Or(
        claripy.And(a, b),
        claripy.Or(claripy.And(a, claripy.Not(b)), claripy.And(claripy.Not(a), c)),
        claripy.Or(claripy.And(a, claripy.Not(b)), claripy.And(claripy.Not(a), claripy.Not(c))),
    )
    assert_correct(complex_true_expression, claripy.true)


def test_simplification():
    def assert_correct(a, b):
        assert claripy.backends.z3.identical(a, b)

    x, y, z = (claripy.BVS(name, 32) for name in ("x", "y", "z"))

    # test extraction of concatted values
    concatted = claripy.Concat(x, y, z)

    assert_correct(concatted[95:64], x)
    assert_correct(concatted[63:32], y)
    assert_correct(concatted[31:0], z)

    assert_correct(concatted[95:32], claripy.Concat(x, y))
    assert_correct(concatted[63:0], claripy.Concat(y, z))

    assert_correct(concatted[95:0], concatted)

    assert_correct(concatted[47:0], claripy.Concat(y, z)[47:0])
    assert_correct(concatted[70:0], concatted[70:0])
    assert_correct(concatted[70:15], concatted[70:15])
    assert_correct(concatted[70:35], claripy.Concat(x, y)[38:3])

    # test extraction of nested concats
    concatted_nested = claripy.Concat(claripy.Reverse(claripy.Concat(x, y)), z)
    assert_correct(concatted_nested[63:0], claripy.Concat(claripy.Reverse(x), z))

    # make sure the division simplification works
    assert_correct(2 + x, claripy.backends.z3.simplify(1 + x + 1))
    assert_correct(x / y, claripy.backends.z3.simplify(x / y))
    assert_correct(x % y, claripy.backends.z3.simplify(x % y))


def test_rotate_shift_mask_simplification():
    a = claripy.BVS("N", 32, max=0xC, min=0x1)
    extend_ = claripy.BVS("extend", 32, uninitialized=True)
    a_ext = extend_.concat(a)
    expr = ((a_ext << 3) | (claripy.LShR(a_ext, 61))) & 0x7FFFFFFF8
    # print(expr)
    # print(expr._model_vsa)
    model_vsa = expr._model_vsa
    assert model_vsa.lower_bound == 8
    assert model_vsa.upper_bound == 0x60
    assert model_vsa.cardinality == 12


def test_reverse_extract_reverse_simplification():
    # without the reverse_extract_reverse simplifier, loading dx from rdx will result in the following complicated
    # expression:
    #   Reverse(Extract(63, 48, Reverse(BVS('rdx', 64))))

    a = claripy.BVS("rdx", 64)
    dx = claripy.Reverse(claripy.Extract(63, 48, claripy.Reverse(a)))

    # simplification should have kicked in at this moment
    assert dx.op == "Extract"
    assert dx.args[0] == 15
    assert dx.args[1] == 0
    assert dx.args[2] is a


def test_reverse_concat_reverse_simplification():
    # Reverse(Concat(Reverse(a), Reverse(b))) = Concat(b, a)

    a = claripy.BVS("a", 32)
    b = claripy.BVS("b", 32)
    x = claripy.Reverse(claripy.Concat(claripy.Reverse(a), claripy.Reverse(b)))

    assert x.op == "Concat"
    assert x.args[0] is b
    assert x.args[1] is a


def perf_boolean_and_simplification_0():
    # Create a gigantic And AST with many operands, one variable at a time
    bool_vars = [claripy.BoolS("b%d" % i) for i in range(1500)]
    v = bool_vars[0]
    for i in range(1, len(bool_vars)):
        v = claripy.And(v, bool_vars[i])


def perf_boolean_and_simplification_1():
    # Create a gigantic And AST with many operands, many variables at a time
    bool_vars = [claripy.BoolS("b%d" % i) for i in range(500)]
    v = bool_vars[0]
    for i in range(1, len(bool_vars)):
        if v.op == "And":
            v = claripy.And(*(v.args + (bool_vars[i] == False,)))  # pylint:disable=singleton-comparison
        else:
            v = claripy.And(v, bool_vars[i])


def test_concrete_flatten():
    a = claripy.BVS("a", 32)
    b = a + 10
    c = 10 + b
    d = a + 20
    assert c is d

    # to future test writers or debuggers: whether the answer is b_neg or b is not particularly important
    e = a - 10
    f = e + 20
    b_neg = a - -10
    assert f is b_neg

    g = e - 10
    h = a - 20
    assert g is h

    i = d - 10
    assert i is b


def test_mask_eq_constant():
    # <Bool ((0#48 .. (0x0 .. sim_data_4_31_8[0:0])[15:0]) & 0xffff) == 0x0>

    a = claripy.BVS("sim_data", 8, explicit_name=True)
    expr = (claripy.ZeroExt(48, claripy.Extract(15, 0, claripy.Concat(claripy.BVV(0, 63), a[0:0]))) & 0xFFFF) == 0x0

    assert expr.op == "__eq__"
    assert expr.args[0].op == "Extract"
    assert expr.args[0].args[0] == 0 and expr.args[0].args[1] == 0
    assert expr.args[0].args[2] is a
    assert expr.args[1].op == "BVV" and expr.args[1].args == (0, 1)

    # the highest bit of the mask (0x1fff) is not aligned to 8
    # we want the mask to be BVV(16, 0x1fff) instead of BVV(13, 0x1fff)
    a = claripy.BVS("sim_data", 8, explicit_name=True)
    expr = (claripy.ZeroExt(48, claripy.Extract(15, 0, claripy.Concat(claripy.BVV(0, 63), a[0:0]))) & 0x1FFF) == 0x0

    assert expr.op == "__eq__"
    assert expr.args[0].op == "__and__"
    _, arg1 = expr.args[0].args
    assert arg1.size() == 16
    assert arg1.args[0] == 0x1FFF


def test_and_mask_comparing_against_constant_simplifier():
    # A & mask == b  ==>  Extract(_, _, A) == Extract(_, _, b) iff high bits of a and b are zeros
    a = claripy.BVS("a", 8)
    b = claripy.BVV(0x10, 32)

    expr = claripy.ZeroExt(24, a) & 0xFFFF == b
    assert expr is (a == 16)

    expr = claripy.Concat(claripy.BVV(0, 24), a) & 0xFFFF == b
    assert expr is (a == 16)

    # A & mask != b ==> Extract(_, _, A) != Extract(_, _, b) iff high bits of a and b are zeros
    a = claripy.BVS("a", 8)
    b = claripy.BVV(0x102000AA, 32)

    expr = claripy.ZeroExt(24, a) & 0xFFFF == b
    assert expr.is_false()

    expr = claripy.Concat(claripy.BVV(0, 24), a) & 0xFFFF == b
    assert expr.is_false()

    # A & 0 == 0 ==> true
    a = claripy.BVS("a", 32)
    b = claripy.BVV(0, 32)
    expr = (a & 0) == b
    assert expr.is_true()
    expr = (a & 0) == claripy.BVV(1, 32)
    assert expr.is_false()


def test_zeroext_extract_comparing_against_constant_simplifier():
    a = claripy.BVS("a", 8, explicit_name=True)
    b = claripy.BVV(0x28, 16)

    expr = claripy.Extract(15, 0, claripy.ZeroExt(24, a)) == b
    assert expr is (a == claripy.BVV(0x28, 8))

    expr = claripy.Extract(7, 0, claripy.ZeroExt(24, a)) == claripy.BVV(0x28, 8)
    assert expr is (a == claripy.BVV(0x28, 8))

    expr = claripy.Extract(7, 0, claripy.ZeroExt(1, a)) == claripy.BVV(0x28, 8)
    assert expr is (a == claripy.BVV(0x28, 8))

    expr = claripy.Extract(6, 0, claripy.ZeroExt(24, a)) == claripy.BVV(0x28, 7)
    assert expr.op == "__eq__"
    assert expr.args[0].op == "Extract" and expr.args[0].args[0] == 6 and expr.args[0].args[1] == 0
    assert expr.args[0].args[2] is a
    assert expr.args[1].args == (0x28, 7)

    expr = claripy.Extract(15, 0, claripy.Concat(claripy.BVV(0, 48), a)) == b
    assert expr is (a == claripy.BVV(0x28, 8))

    bb = claripy.BVV(0x28, 24)
    d = claripy.BVS("d", 8, explicit_name=True)
    expr = claripy.Extract(23, 0, claripy.Concat(claripy.BVV(0, 24), d)) == bb
    assert expr is (d == claripy.BVV(0x28, 8))

    dd = claripy.BVS("dd", 23, explicit_name=True)
    expr = claripy.Extract(23, 0, claripy.Concat(claripy.BVV(0, 2), dd)) == bb
    assert expr is (dd == claripy.BVV(0x28, 23))

    # this was incorrect before
    # claripy issue #201
    expr = claripy.Extract(31, 8, claripy.Concat(claripy.BVV(0, 24), dd)) == claripy.BVV(0xFFFF, 24)
    assert expr is not (dd == claripy.BVV(0xFFFF, 23))


def test_one_xor_exp_eq_zero():
    var1 = claripy.FPV(150, claripy.fp.FSORT_DOUBLE)
    var2 = claripy.FPS("test", claripy.fp.FSORT_DOUBLE)
    result = var1 <= var2
    expr = claripy.BVV(1, 1) ^ (claripy.If(result, claripy.BVV(1, 1), claripy.BVV(0, 1))) == claripy.BVV(0, 1)

    assert expr is result


def test_bitwise_and_if():
    e = claripy.BVS("e", 8)
    cond1 = e >= 5
    cond2 = e != 5
    ifcond1 = claripy.If(cond1, claripy.BVV(1, 1), claripy.BVV(0, 1))
    ifcond2 = claripy.If(cond2, claripy.BVV(1, 1), claripy.BVV(0, 1))
    result = claripy.If(e > 5, claripy.BVV(1, 1), claripy.BVV(0, 1))
    assert ifcond1 & ifcond2 is result


def test_invert_if():
    cond = claripy.BoolS("cond")
    expr = ~(claripy.If(cond, claripy.BVV(1, 1), claripy.BVV(0, 1)))
    result = claripy.If(claripy.Not(cond), claripy.BVV(1, 1), claripy.BVV(0, 1))
    assert expr is result


def test_sub_constant():
    expr = claripy.BVS("expr", 32)
    assert (expr - 5 == 0) is (expr == 5)


def test_extract():
    cond = claripy.BoolS("cond")
    expr = claripy.If(cond, claripy.BVV(1, 32), claripy.BVV(0, 32))[0:0]
    result = claripy.If(cond, claripy.BVV(1, 1), claripy.BVV(0, 1))
    assert expr is result

    e = claripy.BVS("e", 32)
    expr2 = (~e)[0:0]  # pylint:disable=unsubscriptable-object
    result2 = ~(e[0:0])
    assert expr2 is result2


def perf():
    import timeit  # pylint:disable=import-outside-toplevel

    print(
        timeit.timeit(
            "perf_boolean_and_simplification_0()",
            number=10,
            setup="from __main__ import perf_boolean_and_simplification_0",
        )
    )
    print(
        timeit.timeit(
            "perf_boolean_and_simplification_1()",
            number=10,
            setup="from __main__ import perf_boolean_and_simplification_1",
        )
    )


if __name__ == "__main__":
    test_simplification()
    test_bool_simplification()
    test_rotate_shift_mask_simplification()
    test_reverse_extract_reverse_simplification()
    test_reverse_concat_reverse_simplification()
    test_concrete_flatten()
    test_mask_eq_constant()
    test_and_mask_comparing_against_constant_simplifier()
    test_zeroext_extract_comparing_against_constant_simplifier()
    test_one_xor_exp_eq_zero()
    test_bitwise_and_if()
    test_invert_if()
    test_sub_constant()
    test_extract()
